from sphinx.application import Sphinx

from sphinxcontrib.autodoc_pydantic.autodoc import (
    PydanticValidatorDocumenter,
    PydanticModelDocumenter,
    PydanticConfigClassDocumenter,
    PydanticFieldDocumenter,
    PydanticSettingsDocumenter
)

from sphinxcontrib.autodoc_pydantic.directives import (
    PydanticField,
    PydanticConfigClass,
    PydanticValidator,
    PydanticModel,
    PydanticSettings
)



def setup(app: Sphinx) -> None:
    stem = "autodoc_pydantic_"

    app.add_config_value(f'{stem}show_config', True, True, bool)
    app.add_config_value(f'{stem}show_validators', True, True, bool)

    app.add_config_value(f'{stem}model_member_order', 'groupwise', True, str)
    app.add_config_value(f'{stem}model_show_schema', True, True, bool)
    app.add_config_value(f'{stem}model_show_config', True, True, bool)
    app.add_config_value(f'{stem}model_show_validators', True, True, bool)
    app.add_config_value(f'{stem}model_show_paramlist', False, True, bool)

    app.add_config_value(f'{stem}validator_replace_signature', True, True, bool)
    app.add_config_value(f'{stem}validator_list_fields', True, True, bool)

    app.add_config_value(f'{stem}field_list_validators', True, True, bool)
    app.add_config_value(f'{stem}field_doc_policy', 'both', True, str)
    app.add_config_value(f'{stem}field_show_constraints', True, True, bool)
    app.add_config_value(f'{stem}field_show_alias', True, True, bool)

    app.add_directive_to_domain("py", "pydantic_field", PydanticField)
    app.add_directive_to_domain("py", "pydantic_model", PydanticModel)
    app.add_directive_to_domain("py", "pydantic_settings", PydanticSettings)
    app.add_directive_to_domain("py", "pydantic_config_class", PydanticConfigClass)
    app.add_directive_to_domain("py", "pydantic_validator", PydanticValidator)

    app.setup_extension('sphinx.ext.autodoc')  # Require autodoc extension
    app.add_autodocumenter(PydanticFieldDocumenter)
    app.add_autodocumenter(PydanticModelDocumenter)
    app.add_autodocumenter(PydanticSettingsDocumenter)
    app.add_autodocumenter(PydanticValidatorDocumenter)
    app.add_autodocumenter(PydanticConfigClassDocumenter)



"""This module contains the autodoc extension classes.

"""
import json
from typing import Any, Optional

import pydantic
from docutils.statemachine import StringList
from pydantic import BaseSettings
from pydantic.schema import get_field_schema_validations
from sphinx.ext.autodoc import MethodDocumenter, ClassDocumenter, \
    AttributeDocumenter
from sphinx.util.inspect import object_description

from sphinxcontrib.autodoc_pydantic.inspection import is_pydantic_model, \
    is_validator_by_name, ModelWrapper

tpl_collapse = """
.. raw:: html

   <p><details>
   <summary><a>Show JSON schema</a></summary>

.. code-block:: json

{}

.. raw:: html

   </details></p>
   
"""


class PydanticFieldDocumenter(AttributeDocumenter):
    objtype = 'pydantic_field'
    directivetype = 'pydantic_field'
    priority = 10 + AttributeDocumenter.priority
    option_spec = dict(AttributeDocumenter.option_spec)
    member_order = 0

    @classmethod
    def can_document_member(cls,
                            member: Any,
                            membername: str,
                            isattr: bool,
                            parent: Any) -> bool:
        """Filter only pydantic fields.

        """

        is_val = super().can_document_member(member, membername, isattr, parent)
        is_parent_model = is_pydantic_model(parent.object)
        return is_val and is_parent_model and isattr

    def add_directive_header(self, sig: str) -> None:
        """Add default value to header.

        """

        super().add_directive_header(sig)

        name = self.objpath[-1]
        wrapper = ModelWrapper(self.parent)
        default = wrapper.get_field_properties_by_name(name).get("default")
        if default is not None:
            value = object_description(default)
            sourcename = self.get_sourcename()
            self.add_line('   :value: ' + value, sourcename)

    def add_content(self,
                    more_content: Optional[StringList],
                    no_docstring: bool = False
                    ) -> None:

        cfg = self.env.config["autodoc_pydantic_field_doc_policy"]
        if cfg in ("docstring", "both"):
            super().add_content(more_content, no_docstring)
        if cfg in ("both", "description"):
            self.add_description()

        if self.env.config["autodoc_pydantic_field_show_constraints"]:
            self.add_constraints()

        if self.env.config["autodoc_pydantic_field_list_validators"]:
            self.add_validators()

    def add_constraints(self):
        """Adds section showing all defined constraints.

        """

        name = self.objpath[-1]
        wrapper = ModelWrapper(self.parent)
        field = wrapper.get_field_object_by_name(name)

        constraints = get_field_schema_validations(field)
        constraints = {key: value for key, value in constraints.items()
                       if key not in {"env_names"}}

        if constraints:
            source_name = self.get_sourcename()
            self.add_line(":Constraints:", source_name)
            for key, value in constraints.items():
                line = f"   - **{key}** = {value}"
                self.add_line(line, source_name)

    def add_description(self):
        """Adds description from schema if present.

        """

        name = self.objpath[-1]
        wrapper = ModelWrapper(self.parent)
        properties = wrapper.get_field_properties_by_name(name)

        description = properties.get("description")
        if description is not None:
            source_name = self.get_sourcename()
            self.add_line(description, source_name)
            self.add_line("", source_name)

    def add_validators(self):
        """Add section with all validators that process this field.

        """

        name = self.objpath[-1]
        wrapper = ModelWrapper(self.parent)

        validators = wrapper.get_validators_for_field(name)
        if validators:
            source_name = self.get_sourcename()
            self.add_line(":Validated by:", source_name)
            for validator in validators:
                line = f"   - :py:obj:`{validator.name} <{validator.ref}>`"
                self.add_line(line, source_name)


class PydanticConfigClassDocumenter(ClassDocumenter):
    """Represents specialized Documenter subclass for pydantic model
    configuration.

    """

    objtype = 'pydantic_config_class'
    directivetype = 'pydantic_config_class'
    member_order = 100
    priority = 10 + ClassDocumenter.priority

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)
        self.options["undoc-members"] = True

    @classmethod
    def can_document_member(cls,
                            member: Any,
                            membername: str,
                            isattr: bool,
                            parent: Any) -> bool:
        """Filter only pydantic model configurations.

        """

        is_val = super().can_document_member(member, membername, isattr, parent)
        is_parent_model = is_pydantic_model(parent.object)
        is_config = membername == "Config"
        is_class = isinstance(member, type)
        return is_val and is_parent_model and is_config and is_class

    def add_content(self,
                    more_content: Optional[StringList],
                    no_docstring: bool = False
                    ) -> None:
        """Optionally show model configuration or not.

        """

        if self.env.config["autodoc_pydantic_show_config"]:
            super().add_content(more_content, no_docstring)

    def add_directive_header(self, sig: str) -> None:
        """Optionally show model configuration or not.

        """

        if self.env.config["autodoc_pydantic_show_config"]:
            super().add_directive_header(sig)


class PydanticModelDocumenter(ClassDocumenter):
    """Represents specialized Documenter subclass for pydantic models.

    """

    objtype = 'pydantic_model'
    directivetype = 'pydantic_model'
    priority = 10 + ClassDocumenter.priority

    def __init__(self, *args: Any) -> None:
        super().__init__(*args)
        member_order = self.env.config["autodoc_pydantic_model_member_order"]
        self.options["member-order"] = member_order

    @classmethod
    def can_document_member(cls,
                            member: Any,
                            membername: str,
                            isattr: bool,
                            parent: Any) -> bool:
        """Filter only pydantic models.

        """

        is_val = super().can_document_member(member, membername, isattr, parent)
        is_model = is_pydantic_model(member)
        return is_val and is_model

    def format_signature(self, **kwargs) -> str:
        if not self.env.config["autodoc_pydantic_model_show_paramlist"]:
            return ""
        else:
            return super().format_signature(**kwargs)

    def add_content(self,
                    more_content: Optional[StringList],
                    no_docstring: bool = False
                    ) -> None:
        super().add_content(more_content, no_docstring)

        if self.env.config["autodoc_pydantic_model_show_schema"]:
            self.add_collapsable_schema()

        if self.env.config["autodoc_pydantic_model_show_config"]:
            self.add_config()

        if self.env.config["autodoc_pydantic_model_show_validators"]:
            self.add_validators()

    def add_collapsable_schema(self):

        schema_json = self.object.schema_json()
        schema = json.dumps(json.loads(schema_json), default=str, indent=3)
        lines = [f"   {line}" for line in schema.split("\n")]
        lines = "\n".join(lines)
        lines = tpl_collapse.format(lines).split("\n")
        source_name = self.get_sourcename()

        for line in lines:
            self.add_line(line, source_name)


    def add_config(self):
        """Adds summary section describing the model configuration.

        """

        cfg = self.object.Config
        is_main_config = cfg is pydantic.main.BaseConfig
        is_setting_config = cfg is pydantic.env_settings.BaseSettings.Config
        if is_main_config or is_setting_config:
            return

        values = {key: getattr(cfg, key) for key in dir(cfg)
                  if not key.startswith("_")}

        source_name = self.get_sourcename()
        self.add_line(":Config:", source_name)
        for name, value in values.items():
            line = f"   - **{name}**: *{type(value).__name__} = {value}*"
            self.add_line(line, source_name)

    def add_validators(self):
        """Adds summary section describing all validators with corresponding
        fields.

        """

        wrapper = ModelWrapper(self.object)
        validators = wrapper.get_named_references_for_validators()
        if validators:
            source_name = self.get_sourcename()
            self.add_line(":Validators:", source_name)

            for validator in validators:
                for field in wrapper.get_fields_for_validator(validator.name):
                    line = (f"   - "
                            f":py:obj:`{validator.name} <{validator.ref}>`"
                            f" » "
                            f":py:obj:`{field.name} <{field.ref}>`")
                    self.add_line(line, source_name)


class PydanticSettingsDocumenter(PydanticModelDocumenter):

    objtype = 'pydantic_settings'
    directivetype = 'pydantic_settings'
    priority = 10 + PydanticModelDocumenter.priority

    @classmethod
    def can_document_member(cls,
                            member: Any,
                            membername: str,
                            isattr: bool,
                            parent: Any) -> bool:
        """Filter only pydantic models.

        """

        is_val = super().can_document_member(member, membername, isattr, parent)
        if is_val:
            return issubclass(member, BaseSettings)
        else:
            return False


class PydanticValidatorDocumenter(MethodDocumenter):
    """Represents specialized Documenter subclass for pydantic validators.

    """

    objtype = 'pydantic_validator'
    directivetype = 'pydantic_validator'
    member_order = 50
    priority = 10 + MethodDocumenter.priority

    @classmethod
    def can_document_member(cls,
                            member: Any,
                            membername: str,
                            isattr: bool,
                            parent: Any) -> bool:
        """Filter only pydantic validators.

        """

        is_val = super().can_document_member(member, membername, isattr, parent)
        is_validator = is_validator_by_name(membername, parent.object)
        return is_val and is_validator

    def add_directive_header(self, sig: str) -> None:
        """Optionally show validator header.

        """

        if self.env.config["autodoc_pydantic_show_validators"]:
            super().add_directive_header(sig)

    def format_args(self, **kwargs: Any) -> str:

        if self.env.config["autodoc_pydantic_validator_replace_signature"]:
            return ''

    def add_content(self,
                    more_content: Optional[StringList],
                    no_docstring: bool = False
                    ) -> None:
        """Optionally show validator content.

        """

        if self.env.config["autodoc_pydantic_show_validators"]:
            super().add_content(more_content, no_docstring)

            if self.env.config["autodoc_pydantic_validator_list_fields"]:
                self.add_field_list()

    def add_field_list(self):
        """Adds a field list with all fields that are validated by this
        validator.

        """

        wrapper = ModelWrapper(self.parent)
        fields = wrapper.get_fields_for_validator(self.object_name)

        if not fields:
            return

        source_name = self.get_sourcename()
        self.add_line(":Validates:", source_name)

        for field in fields:
            line = f"   - :py:obj:`{field.name} <{field.ref}>`"
            self.add_line(line, source_name)

        self.add_line("", source_name)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['sphinxcontrib', 'sphinxcontrib.autodoc_pydantic']

package_data = \
{'': ['*']}

install_requires = \
['Sphinx>=3.0', 'pydantic>=1.0']

setup_kwargs = {
    'name': 'autodoc-pydantic',
    'version': '0.1.1',
    'description': 'Seamlessly integrate pydantic models in your Sphinx documentation.',
    'long_description': "# autodoc_pydantic\n\n[![PyPI version](https://badge.fury.io/py/autodoc-pydantic.svg)](https://badge.fury.io/py/autodoc-pydantic)\n![Master](https://github.com/mansenfranzen/autodoc_pydantic/actions/workflows/tests.yml/badge.svg)\n![Python](https://img.shields.io/badge/python-3.6+-blue.svg)\n[![Codacy Badge](https://app.codacy.com/project/badge/Grade/30a083d784f245a98a0d5e6857708cc8)](https://www.codacy.com/gh/mansenfranzen/autodoc_pydantic/dashboard?utm_source=github.com&amp;utm_medium=referral&amp;utm_content=mansenfranzen/autodoc_pydantic&amp;utm_campaign=Badge_Grade)\n[![Codacy Badge](https://app.codacy.com/project/badge/Coverage/30a083d784f245a98a0d5e6857708cc8)](https://www.codacy.com/gh/mansenfranzen/autodoc_pydantic/dashboard?utm_source=github.com&utm_medium=referral&utm_content=mansenfranzen/autodoc_pydantic&utm_campaign=Badge_Coverage)\n\nYou love [pydantic](https://pydantic-docs.helpmanual.io/) :heart: and you want to document your models and configuration settings with [sphinx](https://www.sphinx-doc.org/en/master/)? \n\nPerfect, let's go. But wait, sphinx' [autodoc](https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html) does not integrate too well with pydantic models :confused:. \n\nDon't worry - just `pip install autodoc_pydantic` :relaxed:.\n\n## Features\n\n- :speech_balloon: provides default values, alias and constraints for model fields\n- :link: adds references between validators and corresponding fields\n- :page_with_curl: includes collapsable model json schema\n- :surfer: natively integrates with autodoc extension\n- :paperclip: defines explicit pydantic prefixes for models, settings, fields, validators and model config\n- :clipboard: shows summary section for model configuration and validators\n- :eyes: hides overloaded and redundant model class signature\n- :books: sorts fields, validators and model config within models by type\n- 🍀 Supports `pydantic >= 1.0.0` and `sphinx >= 3.4.0`\n\nAll of these addons are completely configurable.\n\n## Installation\n\n1. Install via `pip install autodoc_pydantic`\n2. Add `'sphinxcontrib.autodoc_pydantic'` to the `extensions` list in `conf.py`:\n\n   ```python\n   extensions = ['sphinxcontrib.autodoc_pydantic']\n   ```\n\n3. Configure `autodoc_pydantic` in `conf.py`:\n\n   ```python\n   autodoc_pydantic_field_show_constraints = True\n   autodoc_pydantic_model_show_schema = True\n   ```\n \n## Usage\n\nThe standard `automodule` directive already understands pydantic models that it encounters by default. No more tweaks are required.\n\n```rest\n.. automodule:: package.module\n   :members:\n```\n\nAdditionally, autodoc_pydantic provides specific directives for models, settings, fields, validators and class config:\n\n```rest\n.. autopydantic_model:: package.module.MyModel\n   :members:\n   \n.. autopydantic_settings:: package.module.MySettings\n   :members:\n   \n.. autopydantic_field:: package.module.MyModel.foobar\n\n.. autopydantic_validator:: package.module.MyModel.validator\n\n.. autopydantic_class_config:: package.module.MyModel.Config\n   :members:\n```\n \n## Configuration\n\n### General \n\n- `autodoc_pydantic_show_config` = *True*: By default document `Config` class as class members. If *False*, hides it completely.\n- `autodoc_pydantic_show_validators` = *True*: By default document pydantic validators as class members. If *False*, hides it completely.\n\n### Models / Settings\n\n- `autodoc_pydantic_model_show_schema` = *True*: By default, adds collapsable section including formatted model json schema.\n- `autodoc_pydantic_model_show_config` = *True*: By default, adds model configuration settings to model doc string.\n- `autodoc_pydantic_model_show_validators` = *True*: By default, adds validator -> field mappings to model doc string.\n- `autodoc_pydantic_model_show_paramlist` = *False*: By default, hides overloaded and redundant parameter list from model signature.\n- `autodoc_pydantic_model_member_order` = *'groupwise'*: By default, sorts model members by type to group fields, validators and class config members.\n\n### Fields\n\n- `autodoc_pydantic_field_list_validators` = *True*: By default, lists all validators processing corresponding field.\n- `autodoc_pydantic_field_doc_policy` = *'both'*: By default, show doc string and and field description. If *'description'*, show field description only. If *'docstring'*, show doc string only.  \n- `autodoc_pydantic_field_show_constraints` = *True*: By default, show field constraints (e.g. minimum, maximum etc.).\n- `autodoc_pydantic_field_show_alias` = *True*: By default, show field alias in signature.\n\n### Validators\n\n- `autodoc_pydantic_validator_show_paramlist` = *False*: By default, hides meaningless parameter list from validators.\n- `autodoc_pydantic_validator_replace_retann` = *True*: By default, replaces validators' return annotation with references to processed fields.\n- `autodoc_pydantic_validator_list_fields` = *True*: By default, adds list of references to validators' doc string.\n",
    'author': 'mansenfranzen',
    'author_email': 'franz.woellert@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/mansenfranzen/autodoc_pydantic',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1',
}


setup(**setup_kwargs)



"""This module contains pydantic specific directives.

"""
from typing import Tuple
import importlib

from docutils.nodes import emphasis
from sphinx.addnodes import (
    desc_signature,
    pending_xref,
    desc_annotation
)

from sphinx.domains.python import PyMethod, PyAttribute, PyClasslike
from sphinx.environment import BuildEnvironment

from sphinxcontrib.autodoc_pydantic.inspection import (
    ModelWrapper,
    NamedReference
)


def create_href(text, target, env) -> pending_xref:
    # create the reference node
    options = {'refdoc': env.docname,
               'refdomain': "py",
               'reftype': "obj",
               'reftarget': target}
    refnode = pending_xref(text, **options)
    classes = ['xref', "py", '%s-%s' % ("py", "obj")]
    refnode += emphasis(text, text, classes=classes)
    return refnode


def create_field_href(reference: NamedReference,
                      env: BuildEnvironment) -> pending_xref:
    return create_href(text=reference.name,
                       target=reference.ref,
                       env=env)


class PydanticValidator(PyMethod):
    """Description of a method."""

    def replace_return_node(self, signode: desc_signature):
        """Replaces the return node with references to validated fields.

        """

        # replace nodes
        signode += desc_annotation("", " » ")

        # get imports, names and fields of validator
        module = importlib.import_module(name=signode['module'])
        model_name, validator_name = signode["fullname"].split(".")
        model = getattr(module, model_name)
        wrapper = ModelWrapper.factory(model)
        fields = wrapper.get_fields_for_validator(validator_name)

        # add field reference nodes
        first_field = fields[0]
        signode += create_field_href(first_field, env=self.env)
        for field in fields[1:]:
            signode += desc_annotation("", ", ")
            signode += create_field_href(field, self.env)

    def handle_signature(self, sig: str, signode: desc_signature) -> Tuple[
        str, str]:
        fullname, prefix = super().handle_signature(sig, signode)

        if self.env.config["autodoc_pydantic_validator_replace_signature"]:
            self.replace_return_node(signode)

        return fullname, prefix

    def get_signature_prefix(self, sig: str) -> str:
        return "validator "


class PydanticField(PyAttribute):
    """Description of an attribute."""

    def get_signature_prefix(self, sig: str) -> str:
        return "field "

    def add_alias(self, signode: desc_signature):
        """Replaces the return node with references to validated fields.

        """

        # get imports, names and fields of validator
        module = importlib.import_module(name=signode['module'])
        model_name, field_name = signode["fullname"].split(".")
        model = getattr(module, model_name)
        wrapper = ModelWrapper.factory(model)
        field = wrapper.get_field_object_by_name(field_name)
        alias = field.alias

        if alias != field_name:
            signode += desc_annotation("", f" (alias '{alias}')")

    def handle_signature(self, sig: str, signode: desc_signature) -> Tuple[
        str, str]:
        fullname, prefix = super().handle_signature(sig, signode)

        if self.env.config["autodoc_pydantic_field_show_alias"]:
            self.add_alias(signode)

        return fullname, prefix


class PydanticModel(PyClasslike):
    """Description of an attribute."""

    def get_signature_prefix(self, sig: str) -> str:
        return "pydantic model "


class PydanticSettings(PydanticModel):
    """Description of an attribute."""

    def get_signature_prefix(self, sig: str) -> str:
        return "pydantic settings "


class PydanticConfigClass(PyClasslike):
    """Description of an attribute."""

    def get_signature_prefix(self, sig: str) -> str:
        return "model "
